﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.Linq;
using System.Runtime.ExceptionServices;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Collections.Generic.Extensions;
using OpcLabs.BaseLib.Extensions.Internal;

namespace OpcLabs.UnitTestLib
{
    static public class TestUtilities
    {
        #region Private Types

        [System.Diagnostics.CodeAnalysis.SuppressMessage("ReSharper", "AutoPropertyCanBeMadeGetOnly.Local")]
        private class TestResult
        {
            public TestResult(
                int sequenceNumber,
                [NotNull] string name)
            {
                SequenceNumber = sequenceNumber;
                Name = name;
                Outcome = UnitTestOutcome.Unknown;
            }

            [CanBeNull]
            public ExceptionDispatchInfo ExceptionDispatchInfo { get; set; }

            [NotNull]
            public string Name { get; private set; }

            public UnitTestOutcome Outcome { get; set; }

            public int SequenceNumber { get; private set; }

            public TimeSpan Time { get; set; }
        }

        #endregion



        #region Static Public Methods

        static public void RunTests([NotNull] IEnumerable<Action> tests)
        {
            // ReSharper disable once RedundantArgumentName
            RunTests(TestsToTestMethods(tests));
        }

        static public void RunTests([NotNull] IEnumerable<Action> tests, int breakAt)
        {
            // ReSharper disable once RedundantArgumentName
            RunTests(TestsToTestMethods(tests), breakAt: FormatName(breakAt));
        }

        static public void RunTests([NotNull] IEnumerable<TestMethod> testMethods)
        {
            RunTests(testMethods, new TestRunParameters());
        }

        static public void RunTests(
            [NotNull] IEnumerable<TestMethod> testMethods, 
            [NotNull] TestRunParameters testRunParameters)
        {
            Debug.Assert(testMethods != null);
            Debug.Assert(testRunParameters != null);

            WriteLine();
            WriteLine("Running Tests");
            WriteLine($"Break At: {String.Join(", ", testRunParameters.BreakAtTests)}");
            // IMPROVE: list the maximum outcome counts and other input parameters

            var testResultList = new List<TestResult>();
            var testResultListDictionary = new Dictionary<UnitTestOutcome, List<TestResult>>();
            for (UnitTestOutcome outcome = UnitTestOutcome.Failed; outcome <= UnitTestOutcome.Unknown; outcome++)
                testResultListDictionary.AddNew(outcome, new List<TestResult>());

            int sequenceNumber = 0;
            foreach (TestMethod testMethod in testMethods)
            {
                Debug.Assert(testMethod != null);
                string name = testMethod.Name;

                sequenceNumber++;

                WriteLine();
                WriteLine();
                WriteLine($"Test #{sequenceNumber}: {name}");
                if (testMethod.IgnoreAttribute != null)
                    WriteLine("Ignore");
                // ReSharper disable once PossibleNullReferenceException
                WriteLine($"Test Categories: {String.Join(", ", testMethod.TestCategoryAttributes.Select(attribute => attribute.TestCategories))}");
                int testTimeout = 30*60*1000;   // IMPROVE: have way to change this default form outside
                if (testMethod.TimeoutAttribute != null)
                {
                    testTimeout = testMethod.TimeoutAttribute.Timeout;
                    WriteLine($"Timeout: {testTimeout}");
                }

                var testResult = new TestResult(sequenceNumber, name);
                testResultList.Add(testResult);

                if (testMethod.IgnoreAttribute == null)
                {
                    if (testRunParameters.BreakAtTests.Contains(name))
                    {
                        WriteLine("Debugger Break");
                        Debugger.Break();
                    }
                    WriteLine($"@{DateTime.UtcNow:s} Begin Test");
                    WriteLine("");

                    var stopwatch = new Stopwatch();
                    stopwatch.Start();
                    try
                    {
                        bool result = PerformTest(testMethod, testTimeout);
                        stopwatch.Stop();
                        testResult.Outcome = result ? UnitTestOutcome.Passed : UnitTestOutcome.Timeout;
                    }
                    catch (AssertInconclusiveException assertInconclusiveException)
                    {
                        stopwatch.Stop();
                        DumpException(assertInconclusiveException);
                        testResult.ExceptionDispatchInfo = ExceptionDispatchInfo.Capture(assertInconclusiveException);
                        testResult.Outcome = UnitTestOutcome.Inconclusive;
                    }
                    catch (Exception exception)
                    {
                        stopwatch.Stop();
                        DumpException(exception);
                        testResult.ExceptionDispatchInfo = ExceptionDispatchInfo.Capture(exception);
                        testResult.Outcome = UnitTestOutcome.Failed;
                    }
                    testResult.Time = stopwatch.Elapsed;
                    WriteLine($"End Test #{sequenceNumber}: {testMethod.Name}");
                    WriteLine($"[{testResult.Time:hh\\:mm\\:ss\\.fff}] {testResult.Outcome}");
                }

                List<TestResult> resultsOfOutcomeList = testResultListDictionary[testResult.Outcome];
                Debug.Assert(resultsOfOutcomeList != null);
                resultsOfOutcomeList.Add(testResult);

                if (testRunParameters.MaximumOutcomeCountDictionary.TryGetValue(testResult.Outcome, out int maximumCount))
                {
                    if (resultsOfOutcomeList.Count >= maximumCount)
                    {
                        WriteLine($"Maximum count ({maximumCount}) of '{testResult.Outcome}' outcomes reached");
                        break;
                    }
                }
            }
            int count = sequenceNumber;

            WriteLine();

            WriteLine();
            WriteLine(
                "Sequence #".PadRight(10) + "  " +
                "Test Name".PadRight(88) +
                "Time".PadRight(16) +
                "Outcome");
            WriteLine(new String('-', 130));
            int i = 1;
            foreach (TestResult testResult in testResultList)
            {
                Debug.Assert(testResult != null);
                WriteLine(
                    testResult.SequenceNumber.ToString().PadLeft(10) + "  " +
                    (testResult.Name + " ").PadRight(87, ((i%5 == 0) ? '.' : ' ')) +
                    " " +
                    testResult.Time.ToString(@"hh\:mm\:ss\.fff").PadRight(16) +
                    (HighlightOutcomeAsError(testResult.Outcome) ? $"*{testResult.Outcome}*" : testResult.Outcome.ToString()));
                i++;
            }
            WriteLine();

            WriteLine();
            WriteLine("Outcome".PadRight(16) + "Count");
            WriteLine(new String('-', 26));
            for (UnitTestOutcome outcome = UnitTestOutcome.Failed; outcome <= UnitTestOutcome.Unknown; outcome++)
                // The ":" after the outcome is to make these lines distinguishable in text search from test outcome lines
                // generated by the Microsoft (R) Test Execution Command Line Tool.
                // ReSharper disable once PossibleNullReferenceException
                WriteLine((outcome + ":").PadRight(16) + testResultListDictionary[outcome].Count);
            WriteLine();

            List<TestResult> failedTestResultList = testResultListDictionary[UnitTestOutcome.Failed];
            Debug.Assert(failedTestResultList != null);
            List<TestResult> inconclusiveTestResultList = testResultListDictionary[UnitTestOutcome.Inconclusive];
            Debug.Assert(inconclusiveTestResultList != null);

            WriteLine();
            // ReSharper disable once PossibleNullReferenceException
            WriteLine($"Inconclusive Test(s): {String.Join(", ", inconclusiveTestResultList.Select(FormatReference))}");
            // ReSharper disable once PossibleNullReferenceException
            WriteLine($"Failed Test(s): {String.Join(", ", failedTestResultList.Select(FormatReference))}");
            WriteLine($"{count} Total Test(s)");
            WriteLine();

            // The loop below will, of course, throw not more than once; but we use 'foreach' to easily obtain the 1st entry.
            foreach (TestResult testResult in failedTestResultList)
            {
                Debug.Assert(testResult != null);
                WriteLine();
                WriteConclusion($"FAILED; First Failed Test: {FormatReference(testResult)}");
                ExceptionDispatchInfo exceptionDispatchInfo = testResult.ExceptionDispatchInfo;
                Debug.Assert(exceptionDispatchInfo != null);
                exceptionDispatchInfo.Throw();
            }

            // The loop below will, of course, throw not more than once; but we use 'foreach' to easily obtain the 1st entry.
            foreach (TestResult testResult in inconclusiveTestResultList)
            {
                Debug.Assert(testResult != null);
                WriteLine();
                WriteConclusion($"INCONCLUSIVE; First Inconclusive Test: {FormatReference(testResult)}");
                ExceptionDispatchInfo exceptionDispatchInfo = testResult.ExceptionDispatchInfo;
                Debug.Assert(exceptionDispatchInfo != null);
                exceptionDispatchInfo.Throw();
            }
            WriteConclusion("SUCCESS (Passed)");
        }

        static public void RunTests([NotNull] IEnumerable<TestMethod> testMethods, [NotNull] string breakAt)
        {
            var testRunParameters = new TestRunParameters();
            testRunParameters.BreakAtTests.Add(breakAt);
            RunTests(testMethods, testRunParameters);
        }

        #endregion



        #region Static Private Methods

        // ReSharper disable once SuggestBaseTypeForParameter
        static private void DumpException([NotNull] Exception exception)
        {
            WriteLine(String.Format(CultureInfo.CurrentCulture, "Exception: {0}", exception));
            WriteLine(exception.GetDescendLines());
        }

        [NotNull]
        static private string FormatName(int i)
        {
            Debug.Assert(i >= 1);
            return String.Format(CultureInfo.InvariantCulture, "Test{0}", i);
        }

        [NotNull]
        static private string FormatReference([NotNull] TestResult testResult)
        {
            return $"{testResult.Name} (#{testResult.SequenceNumber})";
        }

        static private bool HighlightOutcomeAsError(UnitTestOutcome outcome)
        {
            return 
                (outcome == UnitTestOutcome.Failed) ||
                (outcome == UnitTestOutcome.Timeout);   // IMPROVE: highlight some others as well?
        }

        static private bool PerformTest([NotNull] TestMethod testMethod, int millisecondsTimeout)
        {
            Action action = testMethod.Test;
            return action.PerformWithAbortOnTimeout(millisecondsTimeout);
        }
        
        [NotNull]
        static private IEnumerable<TestMethod> TestsToTestMethods([NotNull] IEnumerable<Action> tests)
        {
            int i = 1;
            foreach (Action test in tests)
            {
                Debug.Assert(test != null);
                string name = FormatName(i);
                var testMethod = new DelegatedTestMethod(name, test);
                yield return testMethod;
                i++;
            }
        }

        static private void WriteConclusion([NotNull] string value)
        {
            WriteLine(value);
            Console.WriteLine(value);
        }

        static private void WriteLine()
        {
            // ReSharper disable once IntroduceOptionalParameters.Local
            WriteLine("");
        }

        static private void WriteLine([NotNull] string value)
        {
            Trace.WriteLine(String.Format(CultureInfo.CurrentCulture, "#### {0}", value));
        }

        #endregion
    }
}
