﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Globalization;
using System.Reflection;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Execution;
using OpcLabs.BaseLib.Extensions;
using OpcLabs.BaseLib.Extensions.Internal;
using OpcLabs.BaseLib.Reflection.Extensions;
using OpcLabs.BaseLib.Testing;
using OpcLabs.BaseLib.Utilities;

namespace OpcLabs.UnitTestLib.Testers
{
    // IMPROVE: also have ValidParameterTester

    public class ValidValueDataMemberTester : Tester
    {
        private class Worker : MarshalByRefObject
        {
            public void Test(
                [NotNull] Type typeUnderTest,
                [NotNull] MemberInfo dataMember,
                [CanBeNull] object value)
            {
                object instance = null;
                if (dataMember.IsStatic())
                    AssertUtilities.WriteLine("Data member is static.");
                else
                {
                    AssertUtilities.WriteLine("Creating instance...");
                    instance = Activator.CreateInstance(typeUnderTest, nonPublic: true);
                }
                AssertUtilities.WriteLine("Assigning value to a data member...");
                dataMember.SetValue(instance, value, null);
                AssertUtilities.WriteLine("Done.");
            }
        }

        public ValidValueDataMemberTester() 
        {
            TestCategories.Add("DataMembers");
        }

        [NotNull]
        static private TestMethod CreateAssignmentTestMethod(
            [NotNull] Type typeUnderTest,
            [NotNull] MemberInfo dataMember,
            [NotNull] string objectName,
            [NotNull] Func<object> func)
        {
            return new DelegatedTestMethod(
                string.Format(CultureInfo.InvariantCulture, "Assignment_{0}_{1}", dataMember.Name, objectName), 
                () =>
                    {
                        AssertUtilities.WriteLine("Creating value...");
                        object value = func();
                        // ReSharper disable once ConvertIfStatementToConditionalTernaryExpression
                        if (value == null)
                            AssertUtilities.WriteLine("Value = null");
                        else
                        {
                            AssertUtilities.WriteLine(String.Format(CultureInfo.CurrentCulture, "Value of type: {0}", value.GetType()));
                            string s;
                            try
                            {
                                s = value.ToString();
                            }
                            catch (Exception)
                            {
                                s = "*** Exception in ToString()";
                            }
                            AssertUtilities.WriteLine(String.Format(CultureInfo.CurrentCulture, "Value = {0}", s));
                        }
                        if (dataMember.IsStatic())
                        {
#if (NETFRAMEWORK)
                            using (var isolatedApp = new IsolatedApp<Worker>())
                                isolatedApp.Value.Test(typeUnderTest, dataMember, value);
#else
                            AssertUtilities.Inconclusive("Testing static members is not supported under .NET Standard.");
#endif
                        }
                        else
                            new Worker().Test(typeUnderTest, dataMember, value);
                    });
        }

        protected override void Initialize(Type typeUnderTest)
        {
            base.Initialize(typeUnderTest);

            IEnumerable<MemberInfo> dataMembers = TypeUtilities.HighestLevelMembers(typeUnderTest, typeUnderTest.GetDataMembers());
            foreach (MemberInfo dataMember in dataMembers)
            {
                Debug.Assert(dataMember != null);
                if (!ShouldTestDataMember(dataMember))
                    continue;

                Type dataType = dataMember.GetDataType();
                Debug.Assert(dataType != null);

                var singularityCreators = new Dictionary<string, CreatorEntry<ValidValueAttribute>>();
                IEnumerable<ValidValueAttribute> singularityAttributes =
                    dataMember.GetCustomAttributesNotNull<ValidValueAttribute>(inherit: true);
                int i = 1;
                foreach (ValidValueAttribute validValueAttribute in singularityAttributes)
                {
                    Debug.Assert(validValueAttribute != null);
                    string name = string.Format(CultureInfo.InvariantCulture, "_{0}", i);
                    ValidValueAttribute attribute = validValueAttribute;
                    singularityCreators[name] = new CreatorEntry<ValidValueAttribute>(
                        () => attribute.Value, validValueAttribute);
                    i++;
                }

                foreach (KeyValuePair<string, CreatorEntry<ValidValueAttribute>> creator in singularityCreators)
                {
                    string objectName = creator.Key;
                    CreatorEntry<ValidValueAttribute> creatorEntry = creator.Value;
                    Debug.Assert(objectName != null);
                    Debug.Assert(creatorEntry != null);

                    AddTestMethod(CreateAssignmentTestMethod(typeUnderTest, dataMember, objectName, creatorEntry.Creator));
                }

                IEnumerable<ValidValuesAttribute> pluralityAttributes =
                    dataMember.GetCustomAttributesNotNull<ValidValuesAttribute>(inherit: true);
                Dictionary<string, CreatorEntry<ValidValuesAttribute>> pluralityCreators = 
                    GetCreatorEntriesFromAttributes(dataType, pluralityAttributes, allowSubclasses: true);

                foreach (KeyValuePair<string, CreatorEntry<ValidValuesAttribute>> creator in pluralityCreators)
                {
                    string objectName = creator.Key;
                    CreatorEntry<ValidValuesAttribute> creatorEntry = creator.Value;
                    Debug.Assert(objectName != null);
                    Debug.Assert(creatorEntry != null);

                    AddTestMethod(CreateAssignmentTestMethod(typeUnderTest, dataMember, objectName, creatorEntry.Creator));
                }
            }
        }

        protected virtual bool ShouldTestDataMember([NotNull] MemberInfo dataMember)
        {
            Debug.Assert(dataMember != null);
            // ReSharper disable once ConvertIfStatementToReturnStatement
            if (!dataMember.IsSettable())
                return false;
            return true;
        }
    }
}
