﻿using System;
using System.Collections.Generic;
using System.Threading;
using JetBrains.Annotations;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Collections.Generic.Extensions;
using OpcLabs.EasyOpc;
using OpcLabs.EasyOpc.AlarmsAndEvents;
using OpcLabs.EasyOpc.AlarmsAndEvents.Extensions;
using OpcLabs.EasyOpc.AlarmsAndEvents.OperationModel;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.AlarmsAndEvents
{
    partial class EasyAEClientTest
    {
        #region Test Methods

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcAlarmsAndEvents")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void SubscribeEventsTest_Basic()
        {
            SubscribeEventsTest_Basic(
                AELiveTestEnvironment.OpcLabs_KitEventServer.ServerDescriptors.WithProgId, 
                OpcLabs_KitEventServer.SourceIds.Regular_1s,
                useEventPull:false);
        }

        [TestCategory("Live")]
        [TestCategory("OpcClassic")]
        [TestCategory("OpcAlarmsAndEvents")]
        [TestCategory("TimeDomain")]
        [TestMethod]
        public void SubscribeEventsTest_Basic_EventPull()
        {
            SubscribeEventsTest_Basic(
                AELiveTestEnvironment.OpcLabs_KitEventServer.ServerDescriptors.WithProgId,
                OpcLabs_KitEventServer.SourceIds.Regular_1s,
                useEventPull: true);
        }

        #endregion



        protected void SubscribeEventsTest_Basic(
            [NotNull] ServerDescriptor serverDescriptor,
            [NotNull] AENodeDescriptor sourceDescriptor,
            bool useEventPull)
        {
            TestSection.Arrange();
            AssertUtilities.WriteLine(serverDescriptor);
            AssertUtilities.WriteLine(sourceDescriptor);
            AssertUtilities.WriteLine("");

            TestEnvironment.PrecheckServer(serverDescriptor);

            var senderList = new List<object>();
            var argsList = new List<EasyAENotificationEventArgs>();
            int indexSet = -1;
            var completeEvent = new AutoResetEvent(initialState: false);

            void ClientOnNotification(object sender, EasyAENotificationEventArgs args)
            {
                AssertUtilities.WriteLine(args);
                senderList.Add(sender);
                argsList.Add(args);

                if ((indexSet == -1) &&
                    !(args is null) &&
                    ((args.Exception != null) || ((args.EventData != null)/* && !args.EventData.Quality.IsBad*/)))    // simple events do not have quality
                {
                    completeEvent.Set();
                    indexSet = argsList.Count - 1;
                }
            }

            const int notificationRate = 100;
            const int testState = 777777777;

            EasyAEClient client = new EasyAEClient();

            TestSection.Act();
            const int waitTimeout = 8 * 1000;    // 5 seconds was not always enough
            bool completed;
            if (useEventPull)
            {
                client.PullNotificationQueueCapacity = 1000;
                int handle = client.SubscribeEvents(serverDescriptor, sourceDescriptor, notificationRate, testState);
                int startTick = Environment.TickCount;
                int endTick = Environment.TickCount + waitTimeout;
                do
                {
                    EasyAENotificationEventArgs[] argsArray = client.PullMultipleNotifications(
                        int.MaxValue, 
                        Math.Max(endTick - Environment.TickCount, 0));
                    argsArray?.ForAll(args => ClientOnNotification(client, args));
                } while (Environment.TickCount < startTick + waitTimeout);
                completed = completeEvent.WaitOne(0);
                client.UnsubscribeEvents(handle);
            }
            else
            {
                client.Notification += ClientOnNotification;
                int handle = client.SubscribeEvents(serverDescriptor, sourceDescriptor, notificationRate, testState);
                completed = completeEvent.WaitOne(waitTimeout);
                client.UnsubscribeEvents(handle);
                client.Notification -= ClientOnNotification;
            }

            TestSection.Assert();

            int count = argsList.Count;
            AssertUtilities.WriteLine($"count: {count}");
            AssertUtilities.WriteLine($"indexSet: {indexSet}");
            Assert.IsTrue(completed, "Not completed.");

            for (int i = 0; i < count; i++)
            {
                object sender = senderList[i];
                Assert.IsNotNull(sender);
                Assert.AreSame(client, sender);

                EasyAENotificationEventArgs args = argsList[i];
                Assert.IsNotNull(args);

                Assert.AreEqual(testState, args.Arguments.State); // is correct for value types only
                Assert.IsNull(args.Arguments.Callback);

                Assert.IsTrue(args.Arguments.RefreshWhenActive);
                Assert.AreEqual(new AESubscriptionParameters(notificationRate, new AESubscriptionFilter(sourceDescriptor)), args.Arguments.SubscriptionParameters);
                Assert.AreEqual(serverDescriptor, args.Arguments.ServerDescriptor);

                if (!client.SupportsTechnology(serverDescriptor.Technology))
                {
                    Assert.IsFalse(args.Succeeded);
                    Assert.IsInstanceOfType(args.Exception, typeof(NotSupportedException));
                    continue;
                }

                Assert.IsNull(args.Exception);
                Assert.AreEqual(null, args.ErrorId);

                AEEventData actualEventData = args.EventData;
                if (actualEventData is null)
                    continue;

                Assert.IsNotNull(actualEventData);
                AssertUtilities.WriteLine(actualEventData);

                Assert.AreEqual(sourceDescriptor.QualifiedName, actualEventData.QualifiedSourceName);
                // IMPROVE: test more
            }
        }
    }
}
