﻿using System;
using System.Diagnostics;
using System.Reactive;
using System.ServiceModel;
using JetBrains.Annotations;
using Microsoft.ComplexEventProcessing;
using Microsoft.ComplexEventProcessing.Linq;
using Microsoft.ComplexEventProcessing.ManagementService;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.EasyOpc;
using OpcLabs.EasyOpc.AlarmsAndEvents;
using OpcLabs.EasyOpc.AlarmsAndEvents.ComplexEventProcessing;
using OpcLabs.EasyOpc.AlarmsAndEvents.Reactive;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.Legacy.AlarmsAndEvents.ComplexEventProcesing
{
    [TestClass]
    public class UnitTest1
    {
        // ReSharper disable LocalizableElement

        /// <summary>
        /// Incorrectly use OPC-DA server in place of OPC-A&amp;E.
        /// </summary>
        [TestCategory("OpcAlarmsAndEvents")]
        [TestCategory("OpcClassic")]
        [TestCategory("ComplexEventProcessing")]
        [TestMethod]
        public void ConnectionFailureTest_OpcDAServer()
        {
            Test1("OPCLabs.KitServer.2", new AESubscriptionParameters(100));
        }

        [TestCategory("OpcAlarmsAndEvents")]
        [TestCategory("OpcClassic")]
        [TestCategory("ComplexEventProcessing")]
        [TestMethod]
        public void SuccessTest()
        {
            Test1("OPCLabs.KitEventServer.2", new AESubscriptionParameters(100));
        }


        private void Test1(
            [NotNull] ServerDescriptor serverDescriptor, 
            [NotNull] AESubscriptionParameters subscriptionParameters)
        {
            // Create an embedded StreamInsight server
            //using (var server = Server.Create("Default"))
            using (var server = CreateServer())
            {
                Debug.Assert(server != null);

                // Create a local end point for the server embedded in this program
                var managementService = server.CreateManagementService();
                Debug.Assert(managementService != null);
                var host = new ServiceHost(managementService);
                host.AddServiceEndpoint(typeof(IManagementService), new WSHttpBinding(SecurityMode.Message),
                    "http://localhost/MyStreamInsightServer");
                host.Open();

                /* The following entities will be defined and available in the server for other clients:
                 * serverApp
                 * serverSource
                 * serverSink
                 * serverProcess
                 */

                // CREATE a StreamInsight APPLICATION in the server
                var myApp = server.CreateApplication("serverApp");

                // DEFINE a simple SOURCE (returns a point event every second)
                var observable = AENotificationObservable.Create(serverDescriptor, subscriptionParameters);
                var mySource = myApp
                    .DefineObservable(() => observable)
                    .ToPointStreamable(
                        eventArgs => PointEvent.CreateInsert(DateTimeOffset.Now, (AENotificationPayload)eventArgs),
                        AdvanceTimeSettings.StrictlyIncreasingStartTime);

                // DEPLOY the source to the server for clients to use
                mySource.Deploy("serverSource");

                // Compose a QUERY over the source (return events with severity 20 or higher)
                var myQuery = from e in mySource
                              where e.EventDataPayload.Severity >= 20
                              select e;

                // DEFINE a simple observer SINK (writes the value to the server console)
                var mySink = myApp.DefineObserver(() => Observer.Create<AENotificationPayload>(
                    payload => Console.WriteLine("sink_Server..: {0}", payload)));

                // DEPLOY the sink to the server for clients to use
                mySink.Deploy("serverSink");

                // BIND the query to the sink and RUN it
                var binding = myQuery.Bind(mySink);
                Debug.Assert(binding != null);
                using (/*var proc = */binding.Run("serverProcess"))
                {
                    // Wait...
                    Console.WriteLine("----------------------------------------------------------------");
                    Console.WriteLine("MyStreamInsightServer is running...");
                    Console.WriteLine("----------------------------------------------------------------");
                    Console.WriteLine(" ");
                    AssertUtilities.WaitingPhase(30 * 1000);
                    // IMPROVE: test output
                }
                host.Close();
            }
        }

        static private Server CreateServer()
        {
            try
            {
                return Server.Create("Instance1");
            }
            catch (InvalidOperationException exception)
            {
                if (exception.Message.StartsWith("The evaluation period for Microsoft StreamInsight instance 'Instance1' has expired."))
                    Assert.Inconclusive(exception.Message);
                if (exception.Message.StartsWith("The specified StreamInsight instance 'Instance1' could not be found or is not valid for this product."))
                    Assert.Inconclusive(exception.Message);
                throw;
            }
        }

        // ReSharper restore LocalizableElement
    }
}
