﻿
using System;
using System.Collections.Generic;
using System.Text;
using JetBrains.Annotations;
using Microsoft.Extensions.FileProviders;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using OpcLabs.BaseLib.Collections.Specialized;
using OpcLabs.BaseLib.Extensions.FileProviders;
using OpcLabs.BaseLib.Extensions.FileProviders.Internal;
using OpcLabs.BaseLib.Graphs;
using OpcLabs.BaseLib.Graphs.Extensions;
using OpcLabs.BaseLib.OperationModel.Generic;
using OpcLabs.EasyOpc.UA;
using OpcLabs.EasyOpc.UA.AddressSpace;
using OpcLabs.EasyOpc.UA.Engine;
using OpcLabs.EasyOpc.UA.Extensions;
using OpcLabs.EasyOpc.UA.Extensions.FileProviders;
using OpcLabs.EasyOpc.UA.FileTransfer;
using OpcLabs.EasyOpc.UA.FileTransfer.Graphs;
using OpcLabs.EasyOpc.UA.IO.Extensions;
using OpcLabs.EasyOpc.UA.Navigation;
using OpcLabs.EasyOpc.UA.OperationModel;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcLiveTest.UA.FileTransfer
{
    [TestClass]
    public class EasyUAFileTransferClientTest
    {
        #region Public Constants

        // Unified Automation UaNETServer
        public const string EndpointUrlString = "opc.tcp://localhost:48030";

        public const string FileSystemNodeString = "ns=2;s=Demo.Files.FileSystem";

        #endregion



        #region Static Constructor

        static EasyUAFileTransferClientTest()
        {
            EasyUAClient.SharedParameters.EngineParameters.CertificateAcceptancePolicy.TrustedEndpointUrlStrings.Add(EndpointUrlString);
        }

        #endregion



        #region Static Private Fields

        [NotNull] 
        static private readonly UANodeId FileSystemNodeId = new UANodeId(FileSystemNodeString);

        [NotNull]
        static private readonly Random Random = new Random(System.Environment.TickCount);

        #endregion



        #region Public Properties

        [NotNull]
        public UAEndpointDescriptor EndpointDescriptor =
            new UAEndpointDescriptor(EndpointUrlString)
            {
                EndpointSelectionPolicy = new UAEndpointSelectionPolicy { AllowedMessageSecurityModes = UAMessageSecurityModes.SecuritySignAndEncrypt }
            }.WithUserNameIdentity("root", "secret");

        #endregion



        #region Test Methods

        [TestMethod]
        public void BrowseTreeTest()
        {
            var fileTransferClient = new EasyUAFileTransferClient();
            string directoryName = $"TestDirectory-{Random.Next()}";
            AssertUtilities.WriteLine(directoryName);
            UANodeId directoryNodeId = fileTransferClient.CreateDirectory(EndpointDescriptor, FileSystemNodeId, directoryName);

            string fileName = $"TestFile-{Random.Next()}";
            AssertUtilities.WriteLine(fileName);
            UANodeId fileNodeId = fileTransferClient.CreateFile(EndpointDescriptor, directoryNodeId, fileName);

            ITreeTraversable<UAFileTransferNodeTreePosition> treeTraversable = fileTransferClient.BrowseTree(EndpointDescriptor, FileSystemNodeId);
            Assert.IsNotNull(treeTraversable);
            IEnumerable<UAFileTransferNodeTreePosition> enumerable = treeTraversable.ToEnumerable();

            AssertUtilities.WriteLine();
            bool containsFile = false;
            bool containsDirectory = false;
            foreach (UAFileTransferNodeTreePosition fileTransferNodeTreePosition in enumerable)
            {
                AssertUtilities.WriteLine(fileTransferNodeTreePosition);
                Assert.IsNotNull(fileTransferNodeTreePosition);

                if (!fileTransferNodeTreePosition.IsDirectory &&
                    !(fileTransferNodeTreePosition.ParentPosition is null) &&   // can be improved
                    (fileTransferNodeTreePosition.NodeDescriptor.NodeId.Match(fileNodeId)) &&
                    (fileTransferNodeTreePosition.Name == fileName))
                {
                    Assert.IsFalse(containsFile);
                    containsFile = true;
                }

                if (fileTransferNodeTreePosition.IsDirectory &&
                    !(fileTransferNodeTreePosition.ParentPosition is null) &&   // can be improved
                    (fileTransferNodeTreePosition.NodeDescriptor.NodeId.Match(directoryNodeId)) &&
                    (fileTransferNodeTreePosition.Name == directoryName))
                {
                    Assert.IsFalse(containsDirectory);
                    containsDirectory = true;
                }
            }
            Assert.IsTrue(containsFile);
            Assert.IsTrue(containsDirectory);
        }

        [TestMethod]
        public void CopyFileTest()
        {
            var fileTransferClient = new EasyUAFileTransferClient();
            string existingFileName = $"TestFile-{Random.Next()}";
            var existingFileNamedNodeDescriptor = new UANamedNodeDescriptor(FileSystemNodeId, existingFileName);
            fileTransferClient.WriteAllText(EndpointDescriptor, existingFileNamedNodeDescriptor, "test text");
            string newFileName = $"TestFile-{Random.Next()}";
            fileTransferClient.CopyFile(EndpointDescriptor, existingFileNamedNodeDescriptor, 
                new UANamedNodePath(FileSystemNodeId), newFileName);
            var newFileNamedNodeDescriptor = new UANamedNodeDescriptor(FileSystemNodeId, newFileName);
            string actualText = fileTransferClient.ReadAllText(EndpointDescriptor, newFileNamedNodeDescriptor);
            Assert.AreEqual("test text", actualText);
        }

        [TestMethod]
        public void CreateDirectoryAndDeleteTest()
        {
            var fileTransferClient = new EasyUAFileTransferClient();

            UANodeDescriptor fileSystemNodeDescriptor = FileSystemNodeId;

            string directoryName1 = "MyDirectory1-" + Random.Next();
            AssertUtilities.WriteLine($"Creating first directory, '{directoryName1}'...");
            // ReSharper disable once AssignNullToNotNullAttribute
            UANodeId directoryNodeId1 = fileTransferClient.CreateDirectory(EndpointDescriptor, fileSystemNodeDescriptor, directoryName1);
            AssertUtilities.WriteLine($"Node Id of the first directory: {directoryNodeId1}");

            string directoryName2 = "MyDirectory2-" + Random.Next();
            AssertUtilities.WriteLine($"Creating second directory, '{directoryName2}'...");
            UANodeId directoryNodeId2 = fileTransferClient.CreateDirectory(EndpointDescriptor, fileSystemNodeDescriptor, directoryName2);
            Console.WriteLine($"Node Id of the second directory: {directoryNodeId2}");

            string nestedDirectoryName = "MyDirectory3-" + Random.Next();
            AssertUtilities.WriteLine($"Creating nested directory, '{nestedDirectoryName}'...");
            // Note how directoryNodeId2 (a parent directory) is passed as the 2nd argument to the CreateDirectory method.
            UANodeId nestedDirectoryNodeId = fileTransferClient.CreateDirectory(EndpointDescriptor, directoryNodeId2, nestedDirectoryName);
            AssertUtilities.WriteLine($"Node Id of the nested directory: {nestedDirectoryNodeId}");

            AssertUtilities.WriteLine("Deleting the first directory...");
            fileTransferClient.DeleteDirectory(EndpointDescriptor, fileSystemNodeDescriptor, directoryName1);
        }

        [TestMethod]
        public void CreateDirectoryTest()
        {
            var fileTransferClient = new EasyUAFileTransferClient();
            string directoryName1 = $"TestDirectory-{Random.Next()}";
            string directoryName2 = $"TestDirectory-{Random.Next()}";

            var (innerDirectoryNodeDescriptor1, directoryNodeId1) = fileTransferClient.CreateDirectory(
                EndpointDescriptor,
                new UANamedNodePath(FileSystemNodeId, new string[0]),
                directoryName1);
            AssertUtilities.WriteLine(innerDirectoryNodeDescriptor1);
            // ReSharper disable once AssignNullToNotNullAttribute
            AssertUtilities.WriteLine(directoryNodeId1);
            Assert.IsNotNull(innerDirectoryNodeDescriptor1);
            Assert.IsNotNull(directoryNodeId1);

            Assert.IsTrue(fileTransferClient.IsDirectory(EndpointDescriptor, innerDirectoryNodeDescriptor1));
            Assert.IsTrue(fileTransferClient.IsDirectory(EndpointDescriptor, directoryNodeId1));

            var (innerDirectoryNodeDescriptor2, directoryNodeId2) = fileTransferClient.CreateDirectory(
                EndpointDescriptor,
                new UANamedNodePath(FileSystemNodeId, new[] {directoryName1}), 
                directoryName2);
            AssertUtilities.WriteLine(innerDirectoryNodeDescriptor2);
            // ReSharper disable once AssignNullToNotNullAttribute
            AssertUtilities.WriteLine(directoryNodeId2);
            Assert.IsNotNull(innerDirectoryNodeDescriptor2);
            Assert.IsNotNull(directoryNodeId2);
            
            Assert.IsTrue(innerDirectoryNodeDescriptor2.NodeId.Match(directoryNodeId1));

            Assert.IsTrue(fileTransferClient.IsDirectory(EndpointDescriptor, innerDirectoryNodeDescriptor2));
            Assert.IsTrue(fileTransferClient.IsDirectory(EndpointDescriptor, directoryNodeId2));
        }

        [TestMethod]
        public void DeleteDirectoryIfExistsTest_Negative()
        {
            var fileTransferClient = new EasyUAFileTransferClient();
            string directoryName = $"TestDirectory-{Random.Next()}";
            bool result = fileTransferClient.DeleteDirectoryIfExists(EndpointDescriptor, FileSystemNodeId, directoryName);
            Assert.IsFalse(result);
            //Assert.IsFalse(fileTransferClient.IsDirectory(EndpointDescriptor, directoryName));    // cannot assert this - caching allows false positives
        }

        [TestMethod]
        public void DeleteDirectoryIfExistsTest_Positive()
        {
            var fileTransferClient = new EasyUAFileTransferClient();
            string directoryName = $"TestDirectory-{Random.Next()}";
            UANodeId directoryNodeId = fileTransferClient.CreateDirectory(EndpointDescriptor, FileSystemNodeId, directoryName);
            AssertUtilities.WriteLine(directoryNodeId);
            Assert.IsTrue(fileTransferClient.IsDirectory(EndpointDescriptor, directoryNodeId));
            bool result = fileTransferClient.DeleteDirectoryIfExists(EndpointDescriptor, FileSystemNodeId, directoryName);
            Assert.IsTrue(result);
            //Assert.IsFalse(fileTransferClient.IsDirectory(EndpointDescriptor, directoryName));    // cannot assert this - caching allows false positives
        }

        [TestMethod]
        public void DeleteDirectoryTest()
        {
            var fileTransferClient = new EasyUAFileTransferClient();
            string directoryName = $"TestDirectory-{Random.Next()}";
            UANodeId directoryNodeId = fileTransferClient.CreateDirectory(EndpointDescriptor, FileSystemNodeId, directoryName);
            AssertUtilities.WriteLine(directoryNodeId);
            Assert.IsTrue(fileTransferClient.IsDirectory(EndpointDescriptor, directoryNodeId));
            fileTransferClient.DeleteDirectory(EndpointDescriptor, FileSystemNodeId, directoryName);
        }

        [TestMethod]
        public void DeleteIfExistsTest_Negative()
        {
            var fileTransferClient = new EasyUAFileTransferClient();
            string fileName = $"TestFile-{Random.Next()}";
            bool result = fileTransferClient.DeleteIfExists(EndpointDescriptor, FileSystemNodeId, fileName);
            Assert.IsFalse(result);
            //Assert.IsFalse(fileTransferClient.IsFile(EndpointDescriptor, fileName));    // cannot assert this - caching allows false positives
        }

        [TestMethod]
        public void DeleteIfExistsTest_Positive()
        {
            var fileTransferClient = new EasyUAFileTransferClient();
            string fileName = $"TestFile-{Random.Next()}";
            UANodeId fileNodeId = fileTransferClient.CreateFile(EndpointDescriptor, FileSystemNodeId, fileName);
            AssertUtilities.WriteLine(fileNodeId);
            Assert.IsTrue(fileTransferClient.IsFile(EndpointDescriptor, fileNodeId));
            bool result = fileTransferClient.DeleteIfExists(EndpointDescriptor, FileSystemNodeId, fileName);
            Assert.IsTrue(result);
            //Assert.IsFalse(fileTransferClient.IsFile(EndpointDescriptor, fileName));    // cannot assert this - caching allows false positives
        }

        [TestMethod]
        public void DeleteTest_Directory()
        {
            var fileTransferClient = new EasyUAFileTransferClient();
            string directoryName = $"TestDirectory-{Random.Next()}";
            UANodeId directoryNodeId = fileTransferClient.CreateDirectory(EndpointDescriptor, FileSystemNodeId, directoryName);
            AssertUtilities.WriteLine(directoryNodeId);
            Assert.IsTrue(fileTransferClient.IsDirectory(EndpointDescriptor, directoryNodeId));
            fileTransferClient.Delete(EndpointDescriptor, FileSystemNodeId, existingFileOrDirectoryName:directoryName);
        }

        [TestMethod]
        public void DeleteTest_Directory_NonEmpty()
        {
            var fileTransferClient = new EasyUAFileTransferClient();
            string directoryName1 = $"TestDirectory-{Random.Next()}";
            string directoryName2 = $"TestDirectory-{Random.Next()}";

            var (innerDirectoryNodeDescriptor1, directoryNodeId1) = fileTransferClient.CreateDirectory(
                EndpointDescriptor,
                new UANamedNodePath(FileSystemNodeId, new string[0]),
                directoryName1);
            AssertUtilities.WriteLine(innerDirectoryNodeDescriptor1);
            // ReSharper disable once AssignNullToNotNullAttribute
            AssertUtilities.WriteLine(directoryNodeId1);
            Assert.IsNotNull(innerDirectoryNodeDescriptor1);
            var (innerDirectoryNodeDescriptor2, directoryNodeId2) = fileTransferClient.CreateDirectory(
                EndpointDescriptor,
                new UANamedNodePath(FileSystemNodeId, new[] { directoryName1 }),
                directoryName2);
            AssertUtilities.WriteLine(innerDirectoryNodeDescriptor2);
            // ReSharper disable once AssignNullToNotNullAttribute
            AssertUtilities.WriteLine(directoryNodeId2);

            fileTransferClient.Delete(EndpointDescriptor, innerDirectoryNodeDescriptor1, 
                new string[0], directoryNodeId1);
        }

        [TestMethod]
        public void FindFileSystemTest_Negative()
        {
            var fileTransferClient = new EasyUAFileTransferClient();
            Assert.IsNull(fileTransferClient.FindFileSystem(EndpointDescriptor));
        }

        [TestMethod]
        public void FindFileSystemTest_Positive()
        {
            var fileTransferClient = new EasyUAFileTransferClient();
            UANodeDescriptor fileSystemNodeDescriptor = fileTransferClient.FindFileSystem(EndpointDescriptor, "ns=2;s=Demo.Files");
            AssertUtilities.WriteLine(fileSystemNodeDescriptor);
            Assert.IsNotNull(fileSystemNodeDescriptor);
            //Assert.AreEqual(FileSystemNodeId, fileSystemNodeDescriptor.NodeId);
        }

        [TestMethod]
        public void FindFileTest_Positive()
        {
            var fileTransferClient = new EasyUAFileTransferClient();

            string fileName = $"TestFile-{Random.Next()}";
            UANodeId fileNodeId1 = fileTransferClient.CreateFile(EndpointDescriptor, FileSystemNodeId, fileName);

            UANodeDescriptor fileNodeDescriptor = fileTransferClient.FindFile(EndpointDescriptor,
                new UANamedNodeDescriptor(FileSystemNodeId, fileName));
            Assert.IsNotNull(fileNodeDescriptor);
            Assert.IsTrue(fileNodeDescriptor.NodeId.Match(fileNodeId1));
        }

        [TestMethod]
        public void FindParentDirectoryTest()
        {
            var fileTransferClient = new EasyUAFileTransferClient();
            UANodeId directoryNodeId = fileTransferClient.CreateDirectory(EndpointDescriptor, FileSystemNodeId, $"TestDirectory-{Random.Next()}");
            UANodeDescriptor parentNodeDescriptor = fileTransferClient.FindParentDirectory(EndpointDescriptor, directoryNodeId);
            AssertUtilities.WriteLine(parentNodeDescriptor);
            Assert.IsNotNull(parentNodeDescriptor);
            Assert.IsTrue(parentNodeDescriptor.NodeId.Match(FileSystemNodeId));
        }

        [TestMethod]
        public void FindRootAndPathTest()
        {
            var fileTransferClient = new EasyUAFileTransferClient();
            string directoryName = $"TestDirectory-{Random.Next()}";
            UANodeId directoryNodeId = fileTransferClient.CreateDirectory(EndpointDescriptor, FileSystemNodeId, directoryName);
            string fileName = $"TestFile-{Random.Next()}";
            UANodeId fileNodeId = fileTransferClient.CreateFile(EndpointDescriptor, directoryNodeId, fileName);
            UANamedNodeDescriptor rootNodeDescriptor = fileTransferClient.FindRootAndPath(EndpointDescriptor, fileNodeId);
            AssertUtilities.WriteLine(rootNodeDescriptor);
            Assert.IsNotNull(rootNodeDescriptor);

            UANodeDescriptor finalNodeDescriptor = rootNodeDescriptor.FinalNodeDescriptor;
            Assert.IsNotNull(finalNodeDescriptor);
            Assert.IsNotNull(finalNodeDescriptor.NodeId);

            UANamedNodePath namedNodePath = rootNodeDescriptor.NamedNodePath;
            Assert.IsNotNull(namedNodePath);
            Assert.IsTrue(UANodeId.Match(namedNodePath.BaseNodeDescriptor.NodeId, FileSystemNodeId));

            Assert.AreEqual(1, namedNodePath.PathElements.Count);
            Assert.AreEqual(directoryName, namedNodePath.PathElements[0]);

            Assert.AreEqual(fileName, rootNodeDescriptor.NodeName);
        }

        [TestMethod]
        public void GetDirectoryNameTest()
        {
            var fileTransferClient = new EasyUAFileTransferClient();
            string directoryName = $"TestDirectory-{Random.Next()}";
            UANodeId directoryNodeId = fileTransferClient.CreateDirectory(EndpointDescriptor, FileSystemNodeId, directoryName);
            string actualDirectoryName = fileTransferClient.GetDirectoryName(EndpointDescriptor, directoryNodeId);
            Assert.AreEqual(directoryName, actualDirectoryName);
        }

        [TestMethod]
        public void GetFileNameTest()
        {
            var fileTransferClient = new EasyUAFileTransferClient();
            string fileName = $"TestFile-{Random.Next()}";
            UANodeId fileNodeId = fileTransferClient.CreateFile(EndpointDescriptor, FileSystemNodeId, fileName);
            string actualFileName = fileTransferClient.GetFileName(EndpointDescriptor, fileNodeId);
            Assert.AreEqual(fileName, actualFileName);
        }

        [TestMethod]
        public void GetFileOrDirectoryTest_File()
        {
            var fileTransferClient = new EasyUAFileTransferClient();

            string fileName = $"TestFile-{Random.Next()}";
            UANodeId fileNodeId = fileTransferClient.CreateFile(EndpointDescriptor, FileSystemNodeId, fileName);

            UANodeDescriptor fileOrDirectory = fileTransferClient.GetFileOrDirectory(EndpointDescriptor, 
                new UANamedNodeDescriptor(FileSystemNodeId, fileName));
            Assert.IsNotNull(fileOrDirectory);
            Assert.IsTrue(fileOrDirectory.NodeId.Match(fileNodeId));
        }

        [TestMethod]
        public void GetFileProvider2Test()
        {
            var fileTransferClient = new EasyUAFileTransferClient();

            IFileProvider2 fileProvider2 = fileTransferClient.GetFileProvider2(EndpointDescriptor, "ns=2;s=Demo.Files");
            AssertUtilities.WriteLine(fileProvider2);
            Assert.IsNotNull(fileProvider2);

            IFileInfo fileInfo = fileProvider2.GetFileInfo("");
            AssertUtilities.WriteLine(fileInfo);
            Assert.IsNotNull(fileInfo);
            Assert.IsInstanceOfType(fileInfo, typeof(NotFoundFileInfo2));

            IDirectoryContents directoryContents = fileProvider2.GetDirectoryContents("");
            AssertUtilities.WriteLine(directoryContents);
            Assert.IsNotNull(directoryContents);
            Assert.IsInstanceOfType(directoryContents, typeof(UADirectoryContents));
        }

        [TestMethod]
        public void GetFileSystemTest()
        {
            var fileTransferClient = new EasyUAFileTransferClient();
            UANodeDescriptor fileSystem = fileTransferClient.GetFileSystem(EndpointDescriptor, "ns=2;s=Demo.Files");
            AssertUtilities.WriteLine(fileSystem);
            Assert.IsTrue(fileTransferClient.IsDirectory(EndpointDescriptor, fileSystem));
        }

        [TestMethod]
        public void GetFileSystemTest_Failure()
        {
            var fileTransferClient = new EasyUAFileTransferClient();
            try
            {
                AssertUtilities.WriteLine(fileTransferClient.GetFileSystem(EndpointDescriptor));
            }
            catch (UAException uaException)
            {
                Assert.IsInstanceOfType(uaException.InnerException, typeof(UAInvalidReferenceException));
                return;
            }
            Assert.Fail();
        }

        [TestMethod]
        public void GetMultipleFilePropertiesTest()
        {
            var fileTransferClient = new EasyUAFileTransferClient();

            string fileName1 = $"TestFile-{Random.Next()}";
            DateTime creatingDateTime1 = DateTime.UtcNow;
            UANodeId fileNodeId1 = fileTransferClient.CreateFile(EndpointDescriptor, FileSystemNodeId, fileName1);

            string fileName2 = $"TestFile-{Random.Next()}";
            DateTime creatingDateTime2 = DateTime.UtcNow;
            UANodeId fileNodeId2 = fileTransferClient.CreateFile(EndpointDescriptor, FileSystemNodeId, fileName2);
            fileTransferClient.WriteAllText(EndpointDescriptor, UANamedNodeDescriptor.FromUANodeDescriptor(fileNodeId2), "test text");

            ValueResult<UAFileProperties>[] resultArray = fileTransferClient.GetMultipleFileProperties(EndpointDescriptor,
                new UANodeDescriptor[] {fileNodeId1, fileNodeId2});
            Assert.IsNotNull(resultArray);
            Assert.AreEqual(2, resultArray.Length);

            Assert.IsNotNull(resultArray[0]);
            Assert.IsTrue(resultArray[0].Succeeded);
            UAFileProperties fileProperties1 = resultArray[0].Value;
            AssertUtilities.WriteLine(fileProperties1);
            Assert.IsNotNull(fileProperties1);

            Assert.AreEqual("", fileProperties1.MimeType);
            Assert.AreEqual(0, fileProperties1.OpenCount);
            Assert.AreEqual(0, fileProperties1.Size);
            Assert.IsTrue(fileProperties1.Timestamp >= creatingDateTime1);
            Assert.AreEqual(true, fileProperties1.UserWritable);
            Assert.AreEqual(true, fileProperties1.Writable);

            Assert.IsNotNull(resultArray[1]);
            Assert.IsTrue(resultArray[1].Succeeded);
            UAFileProperties fileProperties2 = resultArray[1].Value;
            AssertUtilities.WriteLine(fileProperties2);
            Assert.IsNotNull(fileProperties2);

            Assert.AreEqual("", fileProperties2.MimeType);
            Assert.AreEqual(0, fileProperties2.OpenCount);
            Assert.AreEqual(9, fileProperties2.Size);
            Assert.IsTrue(fileProperties2.Timestamp >= creatingDateTime2);
            Assert.AreEqual(true, fileProperties2.UserWritable);
            Assert.AreEqual(true, fileProperties2.Writable);
        }

        [TestMethod]
        public void GetMultipleNamesTest()
        {
            var fileTransferClient = new EasyUAFileTransferClient();
            
            string fileName = $"TestFile-{Random.Next()}";
            UANodeId fileNodeId = fileTransferClient.CreateFile(EndpointDescriptor, FileSystemNodeId, fileName);

            string directoryName = $"TestDirectory-{Random.Next()}";
            UANodeId directoryNodeId = fileTransferClient.CreateDirectory(EndpointDescriptor, FileSystemNodeId, directoryName);

            ValueResult<string>[] resultArray = fileTransferClient.GetMultipleNames(EndpointDescriptor, new UANodeDescriptor[] {fileNodeId, directoryNodeId});
            Assert.IsNotNull(resultArray);
            Assert.AreEqual(2, resultArray.Length);

            Assert.IsNotNull(resultArray[0]);
            Assert.IsTrue(resultArray[0].Succeeded);
            Assert.AreEqual(fileName, resultArray[0].Value);

            Assert.IsNotNull(resultArray[1]);
            Assert.IsTrue(resultArray[1].Succeeded);
            Assert.AreEqual(directoryName, resultArray[1].Value);
        }

        [TestMethod]
        public void GetParentDirectoryTest()
        {
            var fileTransferClient = new EasyUAFileTransferClient();
            string directoryName1 = $"TestDirectory-{Random.Next()}";
            UANodeId directoryNodeId1 = fileTransferClient.CreateDirectory(EndpointDescriptor, FileSystemNodeId, directoryName1);
            string directoryName2 = $"TestDirectory-{Random.Next()}";
            UANodeId directoryNodeId2 = fileTransferClient.CreateDirectory(EndpointDescriptor, directoryNodeId1, directoryName2);
            UANodeDescriptor parentDirectory = fileTransferClient.GetParentDirectory(EndpointDescriptor, directoryNodeId2);
            Assert.IsTrue(parentDirectory.NodeId.Match(directoryNodeId1));
        }

        [TestMethod]
        public void GetRootDirectoryTest()
        {
            var fileTransferClient = new EasyUAFileTransferClient();
            string directoryName1 = $"TestDirectory-{Random.Next()}";
            UANodeId directoryNodeId1 = fileTransferClient.CreateDirectory(EndpointDescriptor, FileSystemNodeId, directoryName1);
            string directoryName2 = $"TestDirectory-{Random.Next()}";
            UANodeId directoryNodeId2 = fileTransferClient.CreateDirectory(EndpointDescriptor, directoryNodeId1, directoryName2);
            UANodeDescriptor rootDirectory = fileTransferClient.GetRootDirectory(EndpointDescriptor, directoryNodeId2);
            Assert.IsTrue(rootDirectory.NodeId.Match(FileSystemNodeId));
        }

        [TestMethod]
        public void GetWritableFileProviderTest()
        {
            var fileTransferClient = new EasyUAFileTransferClient();

            IWritableFileProvider writableFileProvider = fileTransferClient.GetWritableFileProvider(EndpointDescriptor, "ns=2;s=Demo.Files");
            AssertUtilities.WriteLine(writableFileProvider);
            Assert.IsNotNull(writableFileProvider);

            IWritableFileInfo writableFileInfo = writableFileProvider.GetWritableFileInfo("");
            AssertUtilities.WriteLine(writableFileInfo);
            Assert.IsNotNull(writableFileInfo);
            Assert.IsInstanceOfType(writableFileInfo, typeof(InvalidWritableFileInfo));

            IWritableDirectoryContents writableDirectoryContents = writableFileProvider.GetWritableDirectoryContents("");
            AssertUtilities.WriteLine(writableDirectoryContents);
            Assert.IsNotNull(writableDirectoryContents);
            Assert.IsInstanceOfType(writableDirectoryContents, typeof(UAWritableDirectoryContents));
        }

        [TestMethod]
        public void HasFileSystemTest_Negative()
        {
            var fileTransferClient = new EasyUAFileTransferClient();
            Assert.IsFalse(fileTransferClient.HasFileSystem(EndpointDescriptor));
        }

        [TestMethod]
        public void HasFileSystemTest_Positive()
        {
            var fileTransferClient = new EasyUAFileTransferClient();
            Assert.IsTrue(fileTransferClient.HasFileSystem(EndpointDescriptor, "ns=2;s=Demo.Files"));
        }

        [TestMethod]
        public void IsDirectoryTest_Name_Positive_Directory()
        {
            var fileTransferClient = new EasyUAFileTransferClient();
            string directoryName = $"TestDirectory-{Random.Next()}";
            fileTransferClient.CreateDirectory(EndpointDescriptor, FileSystemNodeId, directoryName);
            bool isFileOrDirectory = fileTransferClient.IsDirectory(EndpointDescriptor, FileSystemNodeId, directoryName);
            Assert.IsTrue(isFileOrDirectory);
        }

        [TestMethod]
        public void IsFileOrDirectoryTest_Name_Positive_Directory()
        {
            var fileTransferClient = new EasyUAFileTransferClient();
            string directoryName = $"TestDirectory-{Random.Next()}";
            fileTransferClient.CreateDirectory(EndpointDescriptor, FileSystemNodeId, directoryName);
            bool isFileOrDirectory = fileTransferClient.IsFileOrDirectory(EndpointDescriptor, FileSystemNodeId, directoryName);
            Assert.IsTrue(isFileOrDirectory);
        }

        [TestMethod]
        public void IsFileOrDirectoryTest_Negative_NonExistent()
        {
            var fileTransferClient = new EasyUAFileTransferClient();
            bool isFileOrDirectory = fileTransferClient.IsFileOrDirectory(EndpointDescriptor, new UANodeId(1, Guid.Parse("{EFC563F8-D1EE-4883-9162-FC19AE65CCC4}")));
            Assert.IsFalse(isFileOrDirectory);
        }

        [TestMethod]
        public void IsFileOrDirectoryTest_Positive_Directory()
        {
            var fileTransferClient = new EasyUAFileTransferClient();
            string directoryName = $"TestDirectory-{Random.Next()}";
            UANodeId directoryNodeId = fileTransferClient.CreateDirectory(EndpointDescriptor, FileSystemNodeId, directoryName);
            bool isFileOrDirectory = fileTransferClient.IsFileOrDirectory(EndpointDescriptor, directoryNodeId);
            Assert.IsTrue(isFileOrDirectory);
        }

        [TestMethod]
        public void IsFileOrDirectoryTest_Positive_File()
        {
            var fileTransferClient = new EasyUAFileTransferClient();
            string fileName = $"TestFile-{Random.Next()}";
            UANodeId fileNodeId = fileTransferClient.CreateFile(EndpointDescriptor, FileSystemNodeId, fileName);
            bool isFileOrDirectory = fileTransferClient.IsFileOrDirectory(EndpointDescriptor, fileNodeId);
            Assert.IsTrue(isFileOrDirectory);
        }

        [TestMethod]
        public void IsFileTest_Negative_Directory()
        {
            var fileTransferClient = new EasyUAFileTransferClient();
            string directoryName = $"TestDirectory-{Random.Next()}";
            UANodeId directoryNodeId = fileTransferClient.CreateDirectory(EndpointDescriptor, FileSystemNodeId, directoryName);
            bool isFile = fileTransferClient.IsFile(EndpointDescriptor, directoryNodeId);
            Assert.IsFalse(isFile);
        }

        [TestMethod]
        public void IsFileTest_Negative_NonExistent()
        {
            var fileTransferClient = new EasyUAFileTransferClient();
            bool isFile = fileTransferClient.IsFile(EndpointDescriptor, new UANodeId(1, Guid.Parse("{EFC563F8-D1EE-4883-9162-FC19AE65CCC4}")));
            Assert.IsFalse(isFile);
        }

        [TestMethod]
        public void IsFileTest_Positive()
        {
            var fileTransferClient = new EasyUAFileTransferClient();
            string fileName = $"TestFile-{Random.Next()}";
            UANodeId fileNodeId = fileTransferClient.CreateFile(EndpointDescriptor, FileSystemNodeId, fileName);
            bool isFile = fileTransferClient.IsFile(EndpointDescriptor, fileNodeId);
            Assert.IsTrue(isFile);
        }

        [TestMethod]
        public void ListDirectoryNamesTest()
        {
            var fileTransferClient = new EasyUAFileTransferClient();
            string directoryName = $"TestDirectory-{Random.Next()}";
            fileTransferClient.CreateDirectory(EndpointDescriptor, FileSystemNodeId, directoryName);
            StringCollection directoryNames = fileTransferClient.ListDirectoryNames(EndpointDescriptor, FileSystemNodeId);
            Assert.IsNotNull(directoryNames);
            CollectionAssert.Contains(directoryNames, directoryName);
        }

        [TestMethod]
        public void ListFileNamesTest()
        {
            var fileTransferClient = new EasyUAFileTransferClient();
            string fileName = $"TestFile-{Random.Next()}";
            fileTransferClient.CreateFile(EndpointDescriptor, FileSystemNodeId, fileName);
            StringCollection fileNames = fileTransferClient.ListFileNames(EndpointDescriptor, FileSystemNodeId);
            Assert.IsNotNull(fileNames);
            CollectionAssert.Contains(fileNames, fileName);
        }

        [TestMethod]
        public void OpenFileTest()
        {
            var fileTransferClient = new EasyUAFileTransferClient();
            string fileName = $"TestFile-{Random.Next()}";
            fileTransferClient.WriteAllText(EndpointDescriptor, new UANamedNodeDescriptor(FileSystemNodeId, fileName), "test text");
            var (fileNodeDescriptor, fileHandle) = fileTransferClient.OpenFile(EndpointDescriptor, FileSystemNodeId, fileName, UAOpenFileModes.Read);
            Assert.IsNotNull(fileNodeDescriptor);
            Assert.IsNotNull(fileHandle);
            byte[] bytes = fileTransferClient.ReadFile(EndpointDescriptor, fileNodeDescriptor, fileHandle, int.MaxValue);
            fileTransferClient.CloseFile(EndpointDescriptor, fileNodeDescriptor, fileHandle);
            // ReSharper disable once PossibleNullReferenceException
            string text = Encoding.UTF8.GetString(bytes);
            Assert.AreEqual("test text", text);
        }

        [TestMethod]
        public void RenameTest_1()
        {
            var fileTransferClient = new EasyUAFileTransferClient();
            string existingFileName = $"TestFile-{Random.Next()}";
            UANodeId existingFileNodeId = fileTransferClient.CreateFile(EndpointDescriptor, FileSystemNodeId, existingFileName);
            AssertUtilities.WriteLine(existingFileNodeId);
            Assert.IsTrue(fileTransferClient.IsFile(EndpointDescriptor, existingFileNodeId));
            string newFileName = $"TestFile-{Random.Next()}";
            UANodeId newFileNodeId = fileTransferClient.Rename(EndpointDescriptor, FileSystemNodeId, existingFileNodeId, newFileName);
            AssertUtilities.WriteLine(newFileNodeId);
            Assert.IsNotNull(newFileNodeId);
            Assert.IsTrue(fileTransferClient.IsFile(EndpointDescriptor, newFileNodeId));
            Assert.IsFalse(fileTransferClient.IsDirectory(EndpointDescriptor, newFileNodeId));
            //Assert.IsFalse(fileTransferClient.IsFile(EndpointDescriptor, existingFileNodeId));    // cannot assert this - caching allows false positives
        }

        [TestMethod]
        public void RenameTest_2()
        {
            var fileTransferClient = new EasyUAFileTransferClient();
            string existingFileName = $"TestFile-{Random.Next()}";
            UANodeId existingFileNodeId = fileTransferClient.CreateFile(EndpointDescriptor, FileSystemNodeId, existingFileName);
            AssertUtilities.WriteLine(existingFileNodeId);
            Assert.IsTrue(fileTransferClient.IsFile(EndpointDescriptor, existingFileNodeId));
            string newFileName = $"TestFile-{Random.Next()}";
            UANodeId newFileNodeId = fileTransferClient.Rename(EndpointDescriptor, UANamedNodePath.FromUANodeDescriptor(FileSystemNodeId), existingFileName, newFileName);
            AssertUtilities.WriteLine(newFileNodeId);
            Assert.IsNotNull(newFileNodeId);
            Assert.IsTrue(fileTransferClient.IsFile(EndpointDescriptor, newFileNodeId));
            Assert.IsFalse(fileTransferClient.IsDirectory(EndpointDescriptor, newFileNodeId));
            //Assert.IsFalse(fileTransferClient.IsFile(EndpointDescriptor, existingFileNodeId));    // cannot assert this - caching allows false positives
        }

        #endregion
    }
}
