﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Text;
using Microsoft.VisualStudio.TestTools.UnitTesting;
using Opc.Ua;
using OpcLabs.BaseLib.OperationModel;
using OpcLabs.BaseLib.OperationModel.Generic;
using OpcLabs.BaseLib.Timing;
using OpcLabs.EasyOpc.UA.Sdk.PubSub;
using OpcLabs.UAPubSubJson.Sdk.PubSub;
using OpcLabs.UnitTestLib;

namespace OpcLabs.EasyOpcTest.UA.Sdk.PubSub.Json
{
    [TestClass]
    public class JsonNetworkMessageGeneratorTest
    {
        #region Private Types

        class DataSetMessageGeneratorCallbacks 
            : object
            , IDataSetMessageGeneratorCallbacks
            , IJsonDataSetMessageGeneratorCallbacks
        {
            public DataSetMessageGeneratorCallbacks(
                JsonDataSetMessageContentMask dataSetMessageContentMask,
                DataSetFieldContentMask dataSetFieldContentMask,
                StatusCode statusCode)
            {
                DataSetFieldContentMask = dataSetFieldContentMask;
                DataSetMessageContentMask = dataSetMessageContentMask;
                _statusCode = statusCode;
            }

            public ConfigurationVersionDataType ConfigurationVersionData => new ConfigurationVersionDataType();

            public DataSetFieldContentMask DataSetFieldContentMask { get; }

            public ushort DataSetWriterId => 55555;

            public double KeepAlivePeriod => 5.0d;
            
            public uint KeyFrameCount => 1;

            public NamespaceTable GetNamespaceTable() => null;

            public ValueResult<DataSetMetaDataType> ObtainDataSetMetaData()
            {
                var dataSetMetaData = new DataSetMetaDataType();
                return ValueResult.Success(dataSetMetaData);
            }

            public void OnFormatFailure(OperationResult operationResult)
            {
                Assert.Fail();
            }

            public (DateTime, ushort, StatusCode, DataValue[])? PullData()
            {
                DateTime dateTime = new DateTime(2020, 1, 15, 15, 43, 0);
                return (dateTime, 9999, _statusCode, new[]
                {
                    new DataValue(new Variant(1), StatusCodes.Good, dateTime, dateTime)
                        {SourcePicoseconds = 9999, ServerPicoseconds = 9999},
                    new DataValue(new Variant(2), StatusCodes.Uncertain, dateTime, dateTime)
                        {SourcePicoseconds = 9999, ServerPicoseconds = 9999},
                });
            }

            public (DateTime, ushort, StatusCode, DataValue[])? PullEvent()
            {
                Assert.Fail();
                throw new InvalidOperationException();
            }

            public JsonDataSetMessageContentMask DataSetMessageContentMask { get; }

            private readonly StatusCode _statusCode;
        }


        class NetworkMessageGeneratorCallbacks : IJsonNetworkMessageGeneratorCallbacks,
                INetworkMessageGeneratorCallbacks
        {
            public NetworkMessageGeneratorCallbacks(JsonNetworkMessageContentMask networkMessageContentMask)
            {
                NetworkMessageContentMask = networkMessageContentMask;
            }

            public object Origin => new Guid("{3502F1DA-25ED-4AA3-B92B-12BEA32025AC}");

            public object PublisherId => "TestPublisherId";

            public void OnFormatFailure(NetworkMessageGenerator networkMessageGenerator, object source,
                OperationResult operationResult)
            {
                Assert.Fail();
            }

            public JsonNetworkMessageContentMask NetworkMessageContentMask { get; }
        }

        class Observer : IObserver<(byte[], IEnumerable<DataSetMessage>)>
        {
            public void OnCompleted()
            {
                Assert.Fail();
            }

            public void OnError(Exception error)
            {
                Assert.Fail();
            }

            public void OnNext((byte[], IEnumerable<DataSetMessage>) value)
            {
                // ReSharper disable once PossibleNullReferenceException
                var jsonText = Encoding.UTF8.GetString(value.Item1);
                // IMPROVE:
                AssertUtilities.WriteLine();
                Trace.WriteLine(jsonText);
            }
        }

        #endregion



        #region Test Methods

        [TestMethod]
        public void Test_DataSetContent_All()
        {
            Test(
                JsonNetworkMessageContentMask.DataSetMessageHeader,
                JsonDataSetMessageContentMask.DataSetWriterId | 
                    JsonDataSetMessageContentMask.MetaDataVersion |
                    JsonDataSetMessageContentMask.SequenceNumber |
                    JsonDataSetMessageContentMask.Status |
                    JsonDataSetMessageContentMask.Timestamp,
                DataSetFieldContentMask.StatusCode | DataSetFieldContentMask.ServerTimestamp,
                2,
                StatusCodes.Bad);
        }

        [TestMethod]
        public void Test_DataSetContent_None()
        {
            Test(
                JsonNetworkMessageContentMask.DataSetMessageHeader,
                JsonDataSetMessageContentMask.None);
        }

        [TestMethod]
        public void Test_DataSetContent_OnlyDataSetWriterId()
        {
            Test(
                JsonNetworkMessageContentMask.DataSetMessageHeader,
                // ReSharper disable once RedundantArgumentDefaultValue
                JsonDataSetMessageContentMask.DataSetWriterId);
        }

        [TestMethod]
        public void Test_DataSetContent_WithMetaDataVersion()
        {
            Test(
                JsonNetworkMessageContentMask.DataSetMessageHeader, 
                JsonDataSetMessageContentMask.DataSetWriterId | JsonDataSetMessageContentMask.MetaDataVersion);
        }

        [TestMethod]
        public void Test_DataSetContent_WithSequenceNumber()
        {
            Test(
                JsonNetworkMessageContentMask.DataSetMessageHeader,
                JsonDataSetMessageContentMask.DataSetWriterId | JsonDataSetMessageContentMask.SequenceNumber );
        }

        [TestMethod]
        public void Test_DataSetContent_WithStatus()
        {
            Test(
                JsonNetworkMessageContentMask.DataSetMessageHeader,
                JsonDataSetMessageContentMask.DataSetWriterId | JsonDataSetMessageContentMask.Status,
                DataSetFieldContentMask.StatusCode | DataSetFieldContentMask.ServerTimestamp,
                2,
                StatusCodes.Bad);
        }

        [TestMethod]
        public void Test_DataSetContent_WithTimestamp()
        {
            Test(
                JsonNetworkMessageContentMask.DataSetMessageHeader,
                JsonDataSetMessageContentMask.DataSetWriterId | JsonDataSetMessageContentMask.Timestamp);
        }

        [TestMethod]
        public void Test_FieldContent_DataValue_WithStatusCode()
        {
            Test(JsonNetworkMessageContentMask.None, JsonDataSetMessageContentMask.DataSetWriterId, DataSetFieldContentMask.StatusCode, 1);
        }

        [TestMethod]
        public void Test_FieldContent_DataValue_WithServerTimestamp()
        {
            Test(JsonNetworkMessageContentMask.None, JsonDataSetMessageContentMask.DataSetWriterId, DataSetFieldContentMask.ServerTimestamp, 1);
        }

        [TestMethod]
        public void Test_FieldContent_Variant_NonReversible()
        {
            Test(JsonNetworkMessageContentMask.None, JsonDataSetMessageContentMask.DataSetWriterId, DataSetFieldContentMask.RawData, 1);
        }

        [TestMethod]
        public void Test_FieldContent_Variant_Reversible()
        {
            Test(JsonNetworkMessageContentMask.None, JsonDataSetMessageContentMask.DataSetWriterId, DataSetFieldContentMask.None, 1);
        }

        [TestMethod]
        // ReSharper disable IdentifierTypo
        public void Test_Structure_NMH0_SDSM0_DSMH0()
            // ReSharper restore IdentifierTypo
        {
            Test(JsonNetworkMessageContentMask.None);
        }

        [TestMethod]
        // ReSharper disable IdentifierTypo
        public void Test_Structure_NMH0_SDSM0_DSMH1()
            // ReSharper restore IdentifierTypo
        {
            Test(JsonNetworkMessageContentMask.DataSetMessageHeader);
        }

        [TestMethod]
        // ReSharper disable IdentifierTypo
        public void Test_Structure_NMH0_SDSM1_DSMH0()
            // ReSharper restore IdentifierTypo
        {
            Test(JsonNetworkMessageContentMask.SingleDataSetMessage);
        }

        [TestMethod]
        // ReSharper disable IdentifierTypo
        public void Test_Structure_NMH0_SDSM1_DSMH1()
            // ReSharper restore IdentifierTypo
        {
            Test(JsonNetworkMessageContentMask.SingleDataSetMessage |
                 JsonNetworkMessageContentMask.DataSetMessageHeader);
        }

        [TestMethod]
        // ReSharper disable IdentifierTypo
        public void Test_Structure_NMH1_SDSM0_DSMH0()
            // ReSharper restore IdentifierTypo
        {
            Test(JsonNetworkMessageContentMask.NetworkMessageHeader);
        }

        [TestMethod]
        // ReSharper disable IdentifierTypo
        public void Test_Structure_NMH1_SDSM0_DSMH1()
            // ReSharper restore IdentifierTypo
        {
            Test(JsonNetworkMessageContentMask.NetworkMessageHeader |
                 JsonNetworkMessageContentMask.DataSetMessageHeader);
        }

        [TestMethod]
        // ReSharper disable IdentifierTypo
        public void Test_Structure_NMH1_SDSM1_DSMH0()
            // ReSharper restore IdentifierTypo
        {
            Test(JsonNetworkMessageContentMask.NetworkMessageHeader |
                 JsonNetworkMessageContentMask.SingleDataSetMessage);
        }

        [TestMethod]
        // ReSharper disable IdentifierTypo
        public void Test_Structure_NMH1_SDSM1_DSMH1()
            // ReSharper restore IdentifierTypo
        {
            Test(JsonNetworkMessageContentMask.NetworkMessageHeader |
                 JsonNetworkMessageContentMask.SingleDataSetMessage |
                 JsonNetworkMessageContentMask.DataSetMessageHeader);
        }

        #endregion



        #region Private Methods

        private void Test(
            JsonNetworkMessageContentMask jsonNetworkMessageContentMask,
            JsonDataSetMessageContentMask jsonDataSetMessageContentMask = JsonDataSetMessageContentMask.DataSetWriterId,
            DataSetFieldContentMask dataSetFieldContentMask =
                DataSetFieldContentMask.StatusCode | DataSetFieldContentMask.ServerTimestamp,
            int dataSetMessageCount = 2)
        {
            Test(jsonNetworkMessageContentMask, jsonDataSetMessageContentMask, dataSetFieldContentMask, dataSetMessageCount, StatusCodes.Good);
        }

        private void Test(
            JsonNetworkMessageContentMask jsonNetworkMessageContentMask,
            JsonDataSetMessageContentMask jsonDataSetMessageContentMask,
            DataSetFieldContentMask dataSetFieldContentMask,
            int dataSetMessageCount,
            StatusCode statusCode)
        {
            var networkMessageGeneratorCallbacks =
                new NetworkMessageGeneratorCallbacks(jsonNetworkMessageContentMask);
            var jsonNetworkMessageGenerator = new JsonNetworkMessageGenerator(networkMessageGeneratorCallbacks,
                networkMessageGeneratorCallbacks);

            var dataSetMessageGeneratorCallbacks =
                new DataSetMessageGeneratorCallbacks(jsonDataSetMessageContentMask, dataSetFieldContentMask, statusCode);
            for (int i = 0; i < dataSetMessageCount; i++)
            {
                var jsonDataSetMessageGenerator1 = new JsonDataSetMessageGenerator(dataSetMessageGeneratorCallbacks,
                    networkMessageGeneratorCallbacks, dataSetMessageGeneratorCallbacks);
                jsonNetworkMessageGenerator.AddDataSetMessageGenerator(jsonDataSetMessageGenerator1);
            }

            var observer = new Observer();
            jsonNetworkMessageGenerator.Subscribe(observer);
            jsonNetworkMessageGenerator.Generate(SystemTiming.Instance.GetTickCount64());
        }

        #endregion
    }
}
